import json
import os

from ..magics.lineage import lineage_input, extract_lineage
from ..jupyterextensions.authextension import AuthClient
from ..services.clients import DataAccessClient


def add_lineage(read_method):
    def wrapper(self, ns):
        ds = read_method(self, ns)
        if '*' not in ns:
            data_access_client = DataAccessClient(AuthClient.get_access_token, os.environ['DATA_ACCESS_URL'])
            location_response = data_access_client.read_location(ns)
            lineage_input(ds, ns, location_response['version'])
        return ds
    return wrapper


def add_doc_option(write_method):
    def wrapper(self, ns):
        if hasattr(self._df, 'doc'):
            doc = self._df.doc
            # doc can be either str or native json
            if type(doc) is str:
                self.option("dataset-doc", doc)
            else:
                self.option("dataset-doc", json.dumps(doc, indent=2))
        return write_method(self, ns)
    return wrapper


def add_lineage_option(write_method):
    def wrap(self, ns):
        # Create a default dataset lineage or use df.lineage if present
        version = _current_milli_time()
        lineage = extract_lineage(self._df, ns, version)
        if lineage is not None:
            self.option("version", version)
            if type(lineage) is str:
                self.option("lineage-doc", lineage)
            else:
                self.option("lineage-doc", json.dumps(lineage, indent=2))
        write_method(self, ns)
    return wrap


def _current_milli_time():
    import time
    return int(round(time.time() * 1000))



from __future__ import print_function

import json
import os
import time
import warnings

import ipywidgets as widgets
from IPython import get_ipython
from IPython.core.error import UsageError
from IPython.core.magic import (Magics, magics_class, line_magic, cell_magic)
from pyspark.sql import DataFrame

from ..jupyterextensions.authextension import AuthClient
from ..services.clients import DatasetDocClient


def lineage_enabled():
    return get_ipython().find_line_magic(DaplaLineageMagics.lineage.__name__) is not None


def lineage_input(df, path, version):
    """Add lineage info to an input Spark DataFrame."""
    if not isinstance(df, DataFrame):
        raise UsageError("The variable '{}' is not a pyspark DataFrame".format(df))
    if lineage_enabled() and df is not None:
        get_ipython().run_line_magic(DaplaLineageMagics.on_input_load.__name__, "{} {} {}"
                                     .format(path, version, df.schema.json()))


def extract_lineage(df, path, version):
    """Extract lineage info from a given Spark DataFrame."""
    if not isinstance(df, DataFrame):
        raise UsageError("The variable '{}' is not a pyspark DataFrame".format(df))
    elif lineage_enabled():
        try:
            get_ipython().run_line_magic(DaplaLineageMagics.on_output_save.__name__, "{} {} {}"
                                         .format(path, version, df.schema.json()))
            if hasattr(df, 'lineage'):
                return map_lineage(df.lineage)
            # Generate simple lineage
            return get_ipython().run_line_magic(DaplaLineageMagics.lineage_json.__name__, "--path {}".format(path))
        except UsageError:
            # Just skip lineage generation
            return None


def map_lineage(lineage_json):
    def mapper(field):
        if 'selected' in field:
            selections = field['selected']

            def source_filter(source):
                return source['path'] in selections and source['field'] in selections[source['path']]
            return {
                'confidence': 0.99,
                'name': field['name'],
                'sources': list(filter(source_filter, field['sources'])),
                'type': field['type']
            }
        else:
            return {
                'confidence': field['confidence'],
                'name': field['name'],
                'sources': field['sources'],
                'type': field['type']
            }

    return {'lineage': {
        'fields': list(map(mapper, lineage_json['lineage']['fields'])),
        'name': lineage_json['lineage']['name'],
        'sources': lineage_json['lineage']['sources'],
        'type': lineage_json['lineage']['type']
        }
    }

@magics_class
class DaplaLineageMagics(Magics):
    """Magics related to lineage management."""

    def __init__(self, shell, lineage_template_provider):
        # You must call the parent constructor
        super(DaplaLineageMagics, self).__init__(shell)
        self._lineage_template_provider = lineage_template_provider
        self._declared_inputs = {}
        self._declared_outputs = {}
        self._input_trace = {}
        self._output_trace = {}

    @staticmethod
    def display(*objs):
        from IPython.display import display
        display(*objs)

    @staticmethod
    def current_milli_time():
        return int(round(time.time() * 1000))

    @cell_magic
    def input(self, line, cell):
        """
        Register input datasets
        """
        opts, args = self.parse_options(line, '', 'append')
        append = 'append' in opts
        if not append:
            # Reset previous state
            self._declared_inputs = {}
        for line in cell.strip().split('\n'):
            if line.strip().startswith('#'):
                continue
            self._declared_inputs[line] = {}

    @cell_magic
    def output(self, line, cell):
        """
        Register input datasets
        """
        opts, args = self.parse_options(line, '', 'append')
        append = 'append' in opts
        if not append:
            # Reset previous state
            self._declared_outputs = {}
        for line in cell.strip().split('\n'):
            if line.strip().startswith('#'):
                continue
            self._declared_outputs[line] = {}

    @line_magic
    def on_input_load(self, line):
        """
        Internal method. Called by spark decorators after an input dataframe has been loaded.
        """
        path, version, schema = line.split(' ')
        if schema is None:
            warnings.warn('Could not find schema for path {}'.format(path))
            self._input_trace[path] = {}
        else:
            self._input_trace[path] = {"schema": schema, "timestamp": version}
        if path not in self._declared_inputs:
            self.show_missing_declaration_warning(path, self.input.__name__)
        else:
            self._declared_inputs[path] = self._input_trace[path]

    @line_magic
    def on_output_save(self, line):
        """
        Internal method. Called by spark decorators after an output dataframe has been saved.
        """
        path, version, schema = line.split(' ')
        if schema is None:
            warnings.warn('Could not find schema for path {}'.format(path))
            self._output_trace[path] = {}
        else:
            self._output_trace[path] = {"schema": schema, "timestamp": version}
        if path not in self._declared_outputs:
            self.show_missing_declaration_warning(path, self.output.__name__)
        else:
            self._declared_outputs[path] = self._output_trace[path]

    @line_magic
    def lineage_tree(self, line):
        def input_details(ds):
            return '{} ({})'.format(ds[0], 'loaded' if 'schema' in ds[1] else 'not loaded')

        print(u'Input datasets:\n |-- {}'.format('\n |-- '.join(map(input_details, self._declared_inputs.items()))))
        print(u'Output datasets:\n |-- {}'.format('\n |-- '.join(self._declared_outputs.keys())))

    @line_magic
    def lineage_json(self, line):
        opts, args = self.parse_options(line, '', 'path')
        use_path = 'path' in opts
        self.validate_inputs()
        if use_path:
            if args not in self._declared_outputs.keys():
                raise UsageError('Could not find path {} in declared outputs'.format(args))
            return self._lineage_template_provider(self._declared_outputs[args], self._declared_inputs)
        elif not args:
            raise UsageError('Missing dataset name.')
        else:
            ds = self.shell.user_ns[args]
            output_schema = {"schema": ds.schema.json(), "timestamp": self.current_milli_time()}
            return self._lineage_template_provider(output_schema, self._declared_inputs)

    @line_magic
    def lineage(self, line):
        """This line magic assists creating documentation metadata for a given variable_name. The variable_name must
        be resolved to a Spark DataFrame.

        Usage:\\
          %lineage variable_name

        """
        opts, args = self.parse_options(line, '')
        if not args:
            raise UsageError('Missing dataset name.')
        self.validate_inputs()
        try:
            ds = self.shell.user_ns[args]
            # Generate lineage from template
            output_schema = {"schema": ds.schema.json(), "timestamp": self.current_milli_time()}
            ds.lineage = self._lineage_template_provider(output_schema, self._declared_inputs)
        except KeyError:
            raise UsageError("Could not find dataset '{}'".format(args))

        variable_titles = []
        variable_forms = []

        options_layout = widgets.Layout(
            overflow_y='auto',
            overflow_x='hidden',
            min_width='100px',
            max_width='300px',
            max_height='200px',
            flex_flow='column',
            display='flow-root',
            flex_shrink='0'
        )

        def capitalize_with_camelcase(s):
            return s[0].upper() + s[1:]

        for field in ds.lineage['lineage']['fields']:
            variable_titles.append(capitalize_with_camelcase(field['name']))
            options = []
            for key, value in self._declared_inputs.items():
                options.append(widgets.Label(value=key))
                schema_fields = json.loads(value['schema'])['fields']

                primary_options = list(filter(lambda sf: sf['name'] == field['name'], schema_fields))
                additional_options = list(filter(lambda sf: sf['name'] != field['name'], schema_fields))

                for source_field in primary_options:
                    options.append(self.create_checkbox(field, key, source_field, closest_match=True))
                if len(additional_options) > 0:
                    vbox = widgets.VBox(list(map(lambda o: self.create_checkbox(field, key, o), additional_options)),
                                        layout=options_layout)
                    options.append(vbox)

            options_widget = widgets.VBox(options)

            variable_forms.append(widgets.VBox([
                widgets.HTML('<style>.widget-checkbox-label-bold > label > span {font-weight: bold;}</style>'),
                widgets.HTML('Choose one or more sources for the variable <b>{}</b>. '
                             'Closest matching source variables are marked with <b>(*)</b>:'.format(field['name'])),
                options_widget]))

        accordion = widgets.Accordion(children=variable_forms)

        for i in range(len(variable_forms)):
            accordion.set_title(i, variable_titles[i])

        self.display(accordion)

    def create_checkbox(self, field, key, source_field, closest_match=False):
        #closest_match = self.is_closest_match(field['sources'], key, source_field['name'])
        w = widgets.Checkbox(
            description=source_field['name'],
            value=False,
            style={'description_width': '0px'})

        if closest_match:
            w.description = w.description + ' (*)'
            w.add_class('widget-checkbox-label-bold')

        def on_value_change(change):
            if 'selected' not in field:
                field['selected'] = {}
            if key not in field['selected']:
                field['selected'][key] = []
            if change['new']:
                field['selected'][key].append(source_field['name'])
            else:
                field['selected'][key].remove(source_field['name'])
        w.observe(on_value_change, names='value')
        return w

    def show_missing_declaration_warning(self, path, method_ref):
        from IPython.core.display import HTML
        self.display(HTML("<b>Warning:</b> The path {} has not been declared. "
                          "To suppress this warning add the following to a cell in your notebook: "
                          "<pre><code style='background: hsl(220, 80%, 90%)'>%%{}\n{}</code></pre>"
                          .format(path, method_ref, path)))

    def validate_inputs(self):
        if len(self._declared_inputs) == 0:
            raise UsageError("Input datasets are not defined. "
                             "To declare input paths add the following to a cell in your notebook:\n%%{}\n<path>"
                             .format(self.input.__name__))
        for declared_input in self._declared_inputs.items():
            if 'schema' not in declared_input[1]:
                raise UsageError('The dataset with path {} has not been loaded'.format(declared_input[0]))

    @staticmethod
    def is_closest_match(sources, path, field):
        for source in sources:
            if source['field'] == field and source['path'] == path:
                return True
        return False


# In order to actually use these magics, you must register them with a
# running IPython.

def load_ipython_extension(ipython):
    """
    Any module file that define a function named `load_ipython_extension`
    can be loaded via `%load_ext module.path` or be configured to be
    autoloaded by IPython at startup time.
    """
    doc_template_client = DatasetDocClient(AuthClient.get_access_token, os.environ['DOC_TEMPLATE_URL'])
    # This class must be registered with a manually created instance,
    # since its constructor has different arguments from the default:
    magics = DaplaLineageMagics(ipython, doc_template_client.get_lineage_template)
    ipython.register_magics(magics)



import os

from dapla.jupyterextensions.authextension import AuthClient, AuthError
from dapla.services.reader import DataSourceReader
from dapla.services.writer import DataSourceWriter


def read_pandas(path):
    reader = DataSourceReader(AuthClient.get_access_token, os.environ['DATA_ACCESS_URL'],  os.environ['CATALOG_URL'])
    try:
        return reader.read(path)
    except AuthError as err:
        err.print_warning()


def show(path):
    reader = DataSourceReader(AuthClient.get_access_token, os.environ['DATA_ACCESS_URL'],  os.environ['CATALOG_URL'])
    try:
        return reader.list(path)
    except AuthError as err:
        err.print_warning()


def details(path):
    reader = DataSourceReader(AuthClient.get_access_token, os.environ['DATA_ACCESS_URL'],  os.environ['CATALOG_URL'])
    try:
        return reader.details(path)
    except AuthError as err:
        err.print_warning()


def write_pandas(df, path, **option_kwargs):
    writer = DataSourceWriter(
        AuthClient.get_access_token,
        os.environ['DATA_ACCESS_URL'],
        os.environ['METADATA_PUBLISHER_URL'],
        os.environ['METADATA_PUBLISHER_PROJECT_ID'],
        os.environ['METADATA_PUBLISHER_TOPIC_NAME'],
        os.environ['CATALOG_URL']
    )
    try:
        writer.write(df, path, **option_kwargs)
    except AuthError as err:
        err.print_warning()



from .clients import DataAccessClient, DataAccessError, MetadataPublisherClient, CatalogClient
from .gcs import GCSFileSystem
from pathlib import Path
import pyarrow.parquet
import base64
# import utils to register the pyarrow extension types
import pandas.core.arrays._arrow_utils  # noqa

class DataSourceWriter:
    """
    Connect to Dapla's Google Cloud Storage.
    """
    def __init__(
            self,
            user_token_provider,
            data_access_url=None,
            metadata_publisher_url=None,
            metadata_publisher_project_id=None,
            metadata_publisher_topic_name=None,
            catalog_url=None,
    ):
        """
        Args:
            user_token (Optional(str)): The OAuth 2.0 access token. When user_token=None,
                no authentication is performed, and you can only access public data
            data_access_url (str): Endpoint to dapla's data access service
            metadata_publisher_url (str): Endpoint to dapla's data metadata publisher service
            metadata_publisher_project_id (str): Project ID used by metadata publisher service
            metadata_publisher_topic_name (str): Topic name used by metadata publisher service
            catalog_url (str): Endpoint to dapla's data catalog service
        """
        self._data_access_client = DataAccessClient(user_token_provider, data_access_url)
        self._metadata_publisher_client = MetadataPublisherClient(
            user_token_provider, metadata_publisher_url,
            metadata_publisher_project_id, metadata_publisher_topic_name)
        self._catalog_client = CatalogClient(user_token_provider, catalog_url)

    @staticmethod
    def _current_milli_time():
        import time
        return int(round(time.time() * 1000))

    @staticmethod
    def guid():
        from uuid import uuid4
        return uuid4().hex

    def write(self, df, path, **kwargs):
        version = kwargs.pop("version", self._current_milli_time())
        valuation = kwargs.pop("valuation")
        state = kwargs.pop("state")

        location_response = self._data_access_client.write_location(path, version, valuation, state)
        fs, gcs_path = self._get_fs(location_response, path, version)
        pandas.io.parquet.BaseImpl.validate_dataframe(df)

        # Write metadata file
        with fs.open(gcs_path + "/.dataset-meta.json", mode="wb") as buffer:
            valid_metadata = base64.b64decode(location_response["validMetadataJson"])
            buffer.write(valid_metadata)

        # Transfom and write pandas dataframe
        from_pandas_kwargs = {"schema": kwargs.pop("schema", None)}
        #if index is not None:
        #    from_pandas_kwargs["preserve_index"] = index
        table = pyarrow.Table.from_pandas(df, preserve_index=True, **from_pandas_kwargs)
        with fs.open("{}/{}.parquet".format(gcs_path, self.guid()), mode="wb") as buffer:
            pyarrow.parquet.write_table(
                table,
                buffer,
                compression="snappy",
                coerce_timestamps="ms",
                **kwargs,
            )
        # Write metadata signature file
        with fs.open(gcs_path + "/.dataset-meta.json.sign", mode="wb") as buffer:
            valid_metadata = base64.b64decode(location_response["metadataSignature"])
            buffer.write(valid_metadata)

        # Publish metadata signature file created event, this will be used for validation and signals a "commit" of metadata
        self._metadata_publisher_client.data_changed(gcs_path + "/.dataset-meta.json.sign")

        # Update catalog
        self._catalog_client.write_dataset(path, version, valuation, state,
                                           location_response['parentUri'],
                                           location_response["validMetadataJson"],
                                           location_response["metadataSignature"])

    def _get_fs(self, location_response, path, version):
        if not location_response['accessAllowed']:
            raise DataAccessError("Din bruker har ikke tilgang")
        parent_uri = location_response['parentUri']
        if parent_uri.startswith('gs:'):
            gcs_path = "{}{}/{}".format(location_response['parentUri'], path, version)
            fs = GCSFileSystem(location_response['accessToken'], "read_write")
            return fs, gcs_path
        elif parent_uri.startswith('file:'):
            gcs_path = "{}{}/{}".format(parent_uri.lstrip('file:').replace('//', ''), path, version)
            # Create anonymous class to wrap Python's build-in open() function to open a file
            fs = type("", (), dict(open=lambda f, mode: open(f, mode)))
            # Ensure that directory exists
            Path(gcs_path).mkdir(parents=True, exist_ok=True)
            return fs, gcs_path
        else:
            raise DataAccessError("Unknown file scheme: " + parent_uri)



from .clients import DataAccessClient, DataAccessError, CatalogClient
from .gcs import GCSFileSystem
import pyarrow.parquet
# import utils to register the pyarrow extension types
import pandas.core.arrays._arrow_utils  # noqa


class DataSourceReader:
    """
    Connect to Dapla's Google Cloud Storage.
    """
    def __init__(
            self,
            user_token_provider,
            data_access_url=None,
            catalog_url=None,
    ):
        """
        Args:
            user_token (Optional(str)): The OAuth 2.0 access token. When user_token=None,
                no authentication is performed, and you can only access public data
            data_access_url (str): Endpoint to dapla's data access service
        """
        self._data_access_client = DataAccessClient(user_token_provider, data_access_url)
        self._catalog_client = CatalogClient(user_token_provider, catalog_url)

    def list(self, path):
        import pandas as pd
        df = pd.DataFrame(self._catalog_client.list(path)['entries'])
        df['timestamp'] = pd.to_datetime(df['timestamp'], unit='ms')
        return df

    def details(self, path):
        fs, gcs_path = self._get_fs(path)
        import pandas as pd
        return pd.DataFrame(list(map(DataSourceReader._file_item(gcs_path), fs.listdir(gcs_path, True))))

    @staticmethod
    def _file_item(gcs_path):
        return lambda o: {'Size': o['size'], 'Name': DataSourceReader._trimmed_name(gcs_path, o)}

    @staticmethod
    def _trimmed_name(gcs_path, o):
        return o['name'].lstrip(o['bucket']).lstrip(gcs_path)

    def read(self, path, columns=None, to_pandas=True, **kwargs):
        fs, gcs_path = self._get_fs(path)
        parquet_ds = pyarrow.parquet.ParquetDataset(
            gcs_path, filesystem=fs, **kwargs
        )
        if to_pandas:
            kwargs["columns"] = columns
            return parquet_ds.read_pandas(**kwargs).to_pandas(split_blocks=False, self_destruct=True)
        else:
            return fs.read_parquet(gcs_path, columns=columns)

    def _get_fs(self, path):
        # Get full path to latest version in GCS bucket
        location_response = self._data_access_client.read_location(path)
        if not location_response['accessAllowed']:
            raise DataAccessError("Din bruker har ikke tilgang")

        parent_uri = location_response['parentUri']
        if parent_uri.startswith('gs:'):
            fs = GCSFileSystem(location_response['accessToken'], "read_only")
            gcs_path = "{}{}/{}".format(parent_uri, path, location_response['version'])
            return fs, gcs_path
        elif parent_uri.startswith('file:'):
            gcs_path = "{}{}/{}".format(parent_uri.lstrip('file:').replace('//', ''), path, location_response['version'])
            return None, gcs_path
        else:
            raise DataAccessError("Unknown file scheme: " + parent_uri)



"""
Custom GenericOAuthenticator extension
"""
from oauthenticator.generic import GenericOAuthenticator


# Pre-Spawn custom class to retrieve user access token
class EnvGenericOAuthenticator(GenericOAuthenticator):

    async def pre_spawn_start(self, user, spawner):

        self.log.info('Calling pre_spawn_start for: ' + user.name)
        # Retrieve user authentication info from JH
        auth_state = await user.get_auth_state()
        if not auth_state:
            # user has no auth state
            self.log.error('User has no auth state')
            return

        # update env var to pass to notebooks
        self.log.info('Starting notebook for: ' + user.name)

    # Refresh user access and refresh tokens (called periodically)
    async def refresh_user(self, user, handler, force=True):
        import jwt
        import time
        import urllib
        import json
        import os
        from tornado.httpclient import HTTPRequest, AsyncHTTPClient
        from tornado.httputil import url_concat
        # Retrieve user authentication info, decode, and check if refresh is needed
        auth_state = await user.get_auth_state()
        access_token = jwt.decode(auth_state['access_token'], verify=False)
        refresh_token = jwt.decode(auth_state['refresh_token'], verify=False)
        diff_access = access_token['exp']-time.time()
        diff_refresh = refresh_token['exp']-time.time()
        # Allow SPARK_USER_TOKEN_EXPIRY_BUFFER_SECS before expiry
        if diff_access > int(os.environ['SPARK_USER_TOKEN_EXPIRY_BUFFER_SECS']):
            # Access token still valid, function returns True
            refresh_user_return = True
        elif diff_refresh < 0:
            # Refresh token not valid, need to completely reauthenticate
            self.log.info('Refresh token not valid, need to completely reauthenticate for: ' +  user.name)
            refresh_user_return = None
        else:
            # We need to refresh access token (which will also refresh the refresh token)
            self.log.info('Try to refresh user tokens for: ' + user.name)
            refresh_token = auth_state['refresh_token']
            http_client = AsyncHTTPClient()
            url = os.environ.get('OAUTH2_TOKEN_URL')
            params = dict(
                grant_type = 'refresh_token',
                client_id = os.environ.get('OAUTH_CLIENT_ID'),
                client_secret = os.environ.get('OAUTH_CLIENT_SECRET'),
                refresh_token = refresh_token
            )

            headers = {
                "Content-Type": "application/x-www-form-urlencoded"
            }

            req = HTTPRequest(url,
                              method="POST",
                              headers=headers,
                              validate_cert=self.tls_verify,
                              body=urllib.parse.urlencode(params)  # Body is required for a POST...
                              )
            resp = await http_client.fetch(req)

            resp_json = json.loads(resp.body.decode('utf8', 'replace'))

            access_token = resp_json['access_token']
            refresh_token = resp_json.get('refresh_token', None)
            token_type = resp_json['token_type']
            scope = resp_json.get('scope', '')
            if (isinstance(scope, str)):
                scope = scope.split(' ')

            # Determine who the logged in user is
            headers = {
                "Accept": "application/json",
                "User-Agent": "JupyterHub",
                "Authorization": "{} {}".format(token_type, access_token)
            }
            if self.userdata_url:
                url = url_concat(self.userdata_url, self.userdata_params)
            else:
                raise ValueError("Please set the OAUTH2_USERDATA_URL environment variable")

            if self.userdata_token_method == "url":
                url = url_concat(self.userdata_url, dict(access_token=access_token))

            req = HTTPRequest(url,
                              method=self.userdata_method,
                              headers=headers,
                              validate_cert=self.tls_verify,
                              )
            resp = await http_client.fetch(req)
            resp_json = json.loads(resp.body.decode('utf8', 'replace'))

            if not resp_json.get(self.username_key):
                self.log.error("OAuth user contains no key %s: %s", self.username_key, resp_json)
                return

            self.log.info('Updating access and refresh token for: ' + user.name)
            refresh_user_return = {
                'name': resp_json.get(self.username_key),
                'auth_state': {
                    'access_token': access_token,
                    'refresh_token': refresh_token,
                    'oauth_user': resp_json,
                    'scope': scope,
                }
            }
        return refresh_user_return



# This code can be put in any Python module, it does not require IPython
# itself to be running already.  It only creates the magics subclass but
# doesn't instantiate it yet.
from __future__ import print_function
from IPython.core.magic import (Magics, magics_class, line_magic)
from IPython.core.error import UsageError, StdinNotImplementedError
import os
import json
from pyspark.sql import DataFrame
import ipywidgets as widgets
from ..services.clients import DatasetDocClient
from ..jupyterextensions.authextension import AuthClient, AuthError


@magics_class
class DaplaDocumentationMagics(Magics):
    """Magics related to documentation management (loading, saving, editing, ...)."""

    def __init__(self, shell, doc_template_provider):
        # You must call the parent constructor
        super(DaplaDocumentationMagics, self).__init__(shell)
        self._doc_template_provider = doc_template_provider

    @staticmethod
    def ensure_valid_filename(fname):
        if fname == '':
            raise UsageError("Invalid filename '{}'".format(fname))
        if not fname.endswith('.json'):
            fname = fname + ".json"
        return fname

    @staticmethod
    def display(*objs):
        from IPython.display import display
        display(*objs)

    @staticmethod
    def clear_output():
        from IPython.display import clear_output
        clear_output()

    @line_magic
    def document(self, line):
        """This line magic assists creating documentation metadata for a given variable_name. The variable_name must be
        resolved to a Spark DataFrame.

        Usage:\\
          %document [options] variable_name

        Options:

            --nofile  : Documentation is not stored to file
            -f <path> : Specifies a file path where the documentation is stored
        """

        opts, args = self.parse_options(line, 'f:', 'nofile')
        if not args:
            raise UsageError('Missing variable name.')

        fname = opts.get('f')
        use_file_storage = 'nofile' not in opts

        try:
            ds = self.shell.user_ns[args]
        except KeyError:
            raise UsageError("Could not find variable name '{}'".format(args))

        if not isinstance(ds, DataFrame):
            raise UsageError("The variable '{}' is not a pyspark DataFrame".format(args))

        if fname is None and use_file_storage:
            fname = self.shell.ev('input("Enter filename where the documentation should be stored '
                                  '(leave blank for no file)")')
            # Update the user input for future use
            if fname == '':
                contents = "%document --nofile {}".format(line)
            else:
                contents = "%document -f {} {}".format(self.ensure_valid_filename(fname), line)
            self.shell.set_next_input(contents, replace=True)
        elif use_file_storage:
            fname = self.ensure_valid_filename(fname)

        if hasattr(ds, 'doc') and ds.doc is not None:
            try:
                ans = self.shell.ask_yes_no("Do you wish to overwrite existing documentation?", default='n')
            except StdinNotImplementedError:
                # skip execution if this note is run in batch mode
                ans = False

            if ans is False:
                self.clear_output()
                return

        self.clear_output()

        try:
            if use_file_storage:
                file_exists = os.path.isfile(fname)
                if file_exists:
                    with open(fname, 'r') as f:
                        ds.doc = json.load(f)
                else:
                    # Generate doc from template and prepare file
                    ds.doc = self._doc_template_provider(ds.schema.json(), False)
                    with open(fname, 'w', encoding="utf-8") as f:
                        json.dump(ds.doc, f)
            else:
                # Generate doc from template
                ds.doc = self._doc_template_provider(ds.schema.json(), False)
        except AuthError as err:
            err.print_warning()
            return

        variable_titles = []
        variable_forms = []

        form_item_layout = widgets.Layout(
            display='flex',
            flex_flow='row',
            justify_content='space-between'
        )

        def capitalize_with_camelcase(s):
            return s[0].upper() + s[1:]

        for instanceVar in ds.doc['instanceVariables']:
            variable_titles.append(capitalize_with_camelcase(instanceVar['name']))
            form_items = []

            for key in instanceVar.keys():
                if key == 'name':
                    continue
                form_items.append(
                    widgets.Box([widgets.Label(value=capitalize_with_camelcase(key)), self.create_widget(instanceVar, key)],
                                layout=form_item_layout))

            variable_forms.append(widgets.Box(form_items, layout=widgets.Layout(
                display='flex',
                flex_flow='column',
                align_items='stretch',
                width='70%'
            )))

        accordion = widgets.Accordion(children=variable_forms, selected_index=None)

        for i in range(len(variable_forms)):
            accordion.set_title(i, variable_titles[i])

        dataset_doc = widgets.Box([
            widgets.Box([widgets.Label(value='Name'), self.create_widget(ds.doc, 'name')], layout=form_item_layout),
            widgets.Box([widgets.Label(value='Description'), self.create_widget(ds.doc, 'description')],
                        layout=form_item_layout),
            widgets.Box([widgets.Label(value='UnitType'), self.create_widget(ds.doc, 'unitType')],
                        layout=form_item_layout)
            ], layout=widgets.Layout(
                display='flex',
                flex_flow='column',
                align_items='stretch',
                width='70%'
            )
        )

        display_objs = (widgets.HTML('<b style="font-size:14px">Dataset metadata</b>'), dataset_doc,
                        widgets.HTML('<b style="font-size:14px">Instance variables</b>'), accordion)

        def on_button_clicked(b):
            with open(fname, 'w', encoding="utf-8") as f:
                json.dump(ds.doc, f)

        if use_file_storage:
            btn = widgets.Button(description='Save to file', icon='file-code')
            btn.on_click(on_button_clicked)
            out = widgets.Output()
            self.display(widgets.VBox([*display_objs, btn, out]))
        else:
            self.display(widgets.VBox(display_objs))

    def create_widget(self, binding, key):
        if isinstance(binding[key], str):
            return self.create_text_input(binding, key)
        elif isinstance(binding[key], bool):
            return self.create_checkbox_input(binding, key)
        elif isinstance(binding[key], dict) and binding[key].__contains__('enums'):
            return self.create_enum_selector(binding, key)
        elif isinstance(binding[key], dict) and binding[key].__contains__('candidates'):
            return self.create_candidate_selector(binding, key)
        else:
            raise UsageError("Unable to create a widget for '{}' with value '{}'\n{}"
                             .format(key, binding[key], json.dumps(binding, indent=2)))

    def create_text_input(self, binding, key):
        if key == 'description':
            component = widgets.Textarea()
        else:
            component = widgets.Text()

        component.value = binding[key]

        def on_change(v):
            binding[key] = v['new']

        component.observe(on_change, names='value')
        return component

    def create_checkbox_input(self, binding, key):
        component = widgets.Checkbox(indent=False)
        component.value = binding[key]

        def on_change(v):
            binding[key] = v['new']

        component.observe(on_change, names='value')
        return component

    def create_enum_selector(self, binding, key):
        component = widgets.Dropdown()
        component.options = binding[key]['enums']
        component.value = binding[key]['selected-enum']

        def on_change(v):
            binding[key]['selected-enum'] = v['new']

        component.observe(on_change, names='value')
        return component

    def create_candidate_selector(self, binding, key):
        component = widgets.Dropdown()
        component.options = list(map(lambda o: (o['name'], o['id']), binding[key]['candidates']))
        component.value = binding[key]['selected-id']

        def on_change(v):
            binding[key]['selected-id'] = v['new']

        component.observe(on_change, names='value')
        return component


# In order to actually use these magics, you must register them with a
# running IPython.

def load_ipython_extension(ipython):
    """
    Any module file that define a function named `load_ipython_extension`
    can be loaded via `%load_ext module.path` or be configured to be
    autoloaded by IPython at startup time.
    """
    doc_template_client = DatasetDocClient(AuthClient.get_access_token, os.environ['DOC_TEMPLATE_URL'])
    # This class must be registered with a manually created instance,
    # since its constructor has different arguments from the default:
    magics = DaplaDocumentationMagics(ipython, doc_template_client.get_doc_template)
    ipython.register_magics(magics)




import os

"""
From https://github.com/jupyterhub/jupyter-server-proxy/issues/57
Fix the Spark UI link to point to the jupyter-server-proxy
"""

def uiWebUrl(self):
    from urllib.parse import urlparse
    web_url = self._jsc.sc().uiWebUrl().get()
    port = urlparse(web_url).port
    return os.environ['JUPYTERHUB_SERVICE_PREFIX'] + "proxy/{}/jobs/".format(port)



import gcsfs
from google.oauth2.credentials import Credentials


class GCSFileSystem(gcsfs.GCSFileSystem):

    def __init__(self, access_token, access):
        token = Credentials(
            access_token,
            token_uri="https://oauth2.googleapis.com/token",
        )
        super().__init__(token=token, access=access)
        # Temporary bug fix for https://issues.apache.org/jira/browse/ARROW-7867
        # Spark writes an empty file to GCS (to mimic a folder structure) before writing partitioned data
        # Resolve this by ignoring the "empty" file when reading partitioned parquet files
        from pyarrow.parquet import EXCLUDED_PARQUET_PATHS
        from pyarrow.parquet import ParquetManifest
        EXCLUDED_PARQUET_PATHS.add('')
        ParquetManifest._should_silently_exclude = _should_silently_exclude
        # Another bug https://issues.apache.org/jira/browse/ARROW-1644


    def isdir(self, path):
        info = super(gcsfs.GCSFileSystem, self).info(path)
        return info["type"] == "directory"

def _should_silently_exclude(self, file_name):
    from pyarrow.parquet import EXCLUDED_PARQUET_PATHS
    return (file_name.endswith('.crc') or  # Checksums
            file_name.endswith('_$folder$') or  # HDFS directories in S3
            file_name.startswith('.') or  # Hidden files starting with .
            file_name.startswith('_') or  # Hidden files starting with _
            '.tmp' in file_name or  # Temp files
            file_name in EXCLUDED_PARQUET_PATHS)

class GCSConnector:
    """
    Connect to Dapla's Google Cloud Storage.
    """
    def __init__(
            self,
            user_token,
            data_access_url=None,
            catalog_url=None,
            metadata_publisher_url=None,
    ):
        """
        Args:
            user_token (Optional(str)): The OAuth 2.0 access token. When user_token=None,
                no authentication is performed, and you can only access public data
            data_access_url (str): Endpoint to dapla's data access service
            catalog_url (str): Endpoint to dapla's catalog service
            metadata_publisher_url (str): Endpoint to dapla's metadata publisher service
        """
        self._user_token = user_token
        self._data_access_url = data_access_url
        self._catalog_url = catalog_url
        self._metadata_publisher_url = metadata_publisher_url

    def read(self, path):
        if self._user_token is None:
            token = 'anon'
        else:
            token = Credentials(
                None,
                token_uri="https://oauth2.googleapis.com/token",
            )
        fs = gcsfs.GCSFileSystem(token=token)
        return fs.read_parquet(path)





import os
import jwt
import time
from pyspark import SparkContext
from pyspark.sql import DataFrameReader, DataFrameWriter, SparkSession
from ..jupyterextensions.authextension import AuthClient, AuthError
from .decorators import add_lineage, add_lineage_option, add_doc_option

"""
This extension will overload the spark session object (spark) with a method called ``path``.
That means, that the "normal" spark expressions:
>>> spark.read.format("gsim").load("/ns")
and
>>> ds.write.format("gsim").save("/ns")
can be replaced by
>>> spark.read.path("/ns")
and
>>> ds.write.path("/ns")
respectively. 

The ``path`` method will ensure that an access token is (re)loaded (if necessary) and added to the spark context.
"""


def load_extensions():
    DataFrameReader.path = namespace_read
    DataFrameWriter.path = namespace_write


@add_lineage
def namespace_read(self, ns):
    try:
        return get_session().read.format("gsim").load(ns)
    except AuthError as err:
        err.print_warning()


@add_doc_option
@add_lineage_option
def namespace_write(self, ns):
    try:
        self._spark = get_session()
        self.format("gsim").save(ns)
    except AuthError as err:
        err.print_warning()


def get_session():
    session = SparkSession._instantiatedSession
    if should_reload_token(session.sparkContext.getConf()):
        # Fetch new access token
        SparkContext._active_spark_context._conf.set("spark.ssb.access", AuthClient.get_access_token())
    return session


def should_reload_token(conf):
    spark_token = conf.get("spark.ssb.access")
    if spark_token is None:
        # First time fetching the token
        return True

    access_token = jwt.decode(spark_token, verify=False)
    diff_access = access_token['exp'] - time.time()
    # Should fetch new token from server if the access token within the given buffer
    if diff_access > int(os.environ['SPARK_USER_TOKEN_EXPIRY_BUFFER_SECS']):
        return False
    else:
        return True



import os
import requests
from tornado import web
from jupyterhub.handlers import BaseHandler
from jupyterhub.services.auth import HubAuth


class AuthHandler(BaseHandler):
    """
    A custom request handler for JupyterHub. This handler returns user and auth state info
    """

    @web.authenticated
    async def get(self):
        user = await self.get_current_user()
        if user is None:
            self.log.info('User is none')
            # whoami can be accessed via oauth token
            user = self.get_current_user_oauth_token()
        if user is None:
            raise web.HTTPError(403)

        self.log.info('User is ' + user.name)
        auth_state = await user.get_auth_state()
        if not auth_state:
            # user has no auth state
            self.log.error('User has no auth state')
            return

        self.write({
            "username": user.name,
            "access_token": auth_state['access_token'],
            "refresh_token": auth_state['refresh_token'],
        })


class AuthClient:
    """
    A client class that connects to the AuthHandler to retrieve user and auth state info
    """
    @staticmethod
    def get_access_token():
        # Helps getting the correct ssl configs
        hub = HubAuth()
        response = requests.get(os.environ['JUPYTERHUB_HANDLER_CUSTOM_AUTH_URL'],
                                headers={
                                    'Authorization': 'token %s' % hub.api_token
                                }, cert=(hub.certfile, hub.keyfile), verify=hub.client_ca, allow_redirects=False)
        if response.status_code == 200:
            return response.json()['access_token']
        else:
            raise AuthError


class AuthError(Exception):
    """This exception class is used when the communication with the custom auth handler fails.
    This is normally due to stale auth session."""

    def print_warning(self):
        from IPython.core.display import display, HTML
        display(HTML('Your session has timed out. Please <a href="/hub/login">log in</a> to continue.'))



import io
import os
import re

from setuptools import find_packages
from setuptools import setup


def read(filename):
    filename = os.path.join(os.path.dirname(__file__), filename)
    text_type = type(u"")
    with io.open(filename, mode="r", encoding='utf-8') as fd:
        return re.sub(text_type(r':[a-z]+:`~?(.*?)`'), text_type(r'``\1``'), fd.read())


DEPENDENCIES = [
    'ipython>=4.0.2'
    'pyspark'
    'jupyterhub'
    'oauthenticator'
    'jwt'
    'requests'
    'responses'
    'ipykernel>=4.2.2'
    'notebook>=4.2'
    'tornado>=4'
    'gcsfs==0.6.2'
    'pyarrow>=0.17.*'
    'pandas'
    'google-auth>=1.2'
    'google-auth-oauthlib'
    'ipywidgets'
]

setup(
    name="ssb-ipython-kernels",
    version="0.2.6",
    url="https://github.com/statisticsnorway/dapla-ipython-kernels",
    license='MIT',

    author="Statistics Norway",
    author_email="bjorn.skaar@ssb.no",

    description="Jupyter kernels for working with dapla services",
    long_description=read("README.md"),
    long_description_content_type="text/markdown",

    packages=find_packages(exclude=('tests', 'examples',)),

    install_requires=DEPENDENCIES,

    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
)



import requests
import json


class AbstractClient:
    def __init__(
            self,
            user_token_provider,
            base_url,
    ):
        self._user_token_provider = user_token_provider
        self._base_url = base_url.rstrip('/')


class CatalogClient(AbstractClient):

    def list(self, path):
        catalog_url = self._base_url + '/rpc/CatalogService/listByPrefix'
        request = {"prefix": path}
        response = requests.post(catalog_url, json=request,
                                 headers={
                                     'Authorization': 'Bearer %s' % self._user_token_provider()
                                 }, allow_redirects=False)
        handle_error_codes(response)
        return response.json()

    def write_dataset(self, path, version, valuation, state, parentUri, datasetMetaBytes, datasetMetaSignatureBytes):
        catalog_url = self._base_url + '/catalog/write'
        request = {
            "dataset": {
                "id": {
                    "path": path,
                    "timestamp": version
                },
                "valuation": valuation,
                "state": state,
                "parentUri": parentUri
            },
            "datasetMetaBytes": datasetMetaBytes,
            "datasetMetaSignatureBytes": datasetMetaSignatureBytes
        }
        response = requests.post(catalog_url, json=request,
                                 headers={
                                     'Authorization': 'Bearer %s' % self._user_token_provider()
                                 }, allow_redirects=False)
        handle_error_codes(response)


class DataAccessClient(AbstractClient):

    def read_location(self, path):
        data_access_url = self._base_url + '/rpc/DataAccessService/readLocation'
        location_request = {"path": path, "snapshot": 0}
        response = requests.post(data_access_url, json=location_request,
                                 headers={
                                     'Authorization': 'Bearer %s' % self._user_token_provider()
                                 }, allow_redirects=False)
        handle_error_codes(response)
        return response.json()

    def write_location(self, path, version, valuation, state):
        data_access_url = self._base_url + '/rpc/DataAccessService/writeLocation'
        metadata = {
            "id": {
                "path": path,
                "version": version
            },
            "valuation": valuation,
            "state": state
        }
        location_request = {
            "metadataJson": json.dumps(metadata)
        }
        response = requests.post(data_access_url, json=location_request,
                                 headers={
                                     'Authorization': 'Bearer %s' % self._user_token_provider()
                                 }, allow_redirects=False)
        handle_error_codes(response)
        return response.json()


class MetadataPublisherClient(AbstractClient):

    def __init__(
            self,
            user_token_provider,
            base_url,
            project_id,
            topic_name,
    ):
        super().__init__(user_token_provider, base_url)
        self._project_id = project_id
        self._topic_name = topic_name

    def data_changed(self, dataset_uri):
        publisher_url = self._base_url + '/rpc/MetadataDistributorService/dataChanged'
        if dataset_uri.startswith('/'):
            # Metadata publisher uses file scheme
            dataset_uri = 'file://' + dataset_uri
        request = {
            "projectId": self._project_id,
            "topicName": self._topic_name,
            "uri": dataset_uri
        }
        response = requests.post(publisher_url, json=request,
                                 headers={
                                     'Authorization': 'Bearer %s' % self._user_token_provider()
                                 }, allow_redirects=False)
        handle_error_codes(response)


class DatasetDocClient(AbstractClient):

    def get_doc_template(self, spark_schema, use_simple):
        request_url = self._base_url + '/doc/template'
        request = {
            "schema": spark_schema,
            "schemaType": "SPARK",
            "useSimpleFiltering": use_simple
        }
        response = requests.post(request_url, json=request,
                                 headers={
                                     'Authorization': 'Bearer %s' % self._user_token_provider()
                                 }, allow_redirects=False)
        handle_error_codes(response)
        return response.json()

    def get_lineage_template(self, output_schema, input_schema_map, use_simple=False):
        request_url = self._base_url + '/lineage/template'

        def mapper(x):
            return (x[0], {
                "schema": x[1]['schema'],
                "schemaType": "SPARK",
                "timestamp": x[1]['timestamp'],
            })
        request = {
            "schema": output_schema['schema'],
            "timestamp": output_schema['timestamp'],
            "schemaType": "SPARK",
            "simpleLineage": use_simple,
            "dependencies": [dict(map(mapper, input_schema_map.items()))],
        }
        response = requests.post(request_url, json=request,
                                 headers={
                                     'Authorization': 'Bearer %s' % self._user_token_provider()
                                 }, allow_redirects=False)
        handle_error_codes(response)
        return response.json()


def handle_error_codes(response):
    if response.status_code == 401:
        raise DataAccessError("Feil med tilgang til tjenesten", response.text)
    elif response.status_code == 404:
        raise DataAccessError("Fant ikke datasett", response.content)
    elif response.status_code < 200 | response.status_code >= 400:
        raise DataAccessError("En feil har oppstått: {}".format(response), response.text)


class DataAccessError(Exception):
    pass



from .documentation import DaplaDocumentationMagics
from .lineage import DaplaLineageMagics


def load_all(ipython):
    from .documentation import load_ipython_extension
    load_ipython_extension(ipython)
    from .lineage import load_ipython_extension
    load_ipython_extension(ipython)
